import re

from amino import Maybe, List, __
from amino.test import later
from amino.test.path import fixture_path, base_dir

from myo.output.data import OutputEntry, OutputEvent, ParseResult

from integration._support.tmux import TmuxCmdSpec
from integration._support.vimtest import vimtest

_parse_head = 'parsed output:'
_event_head = 'event 1:'


def _parse_echo(data):
    matcher = re.compile('^line (\d+)')
    matches = data / matcher.match // Maybe / __.group(1)
    entries = matches / (lambda a: OutputEntry(text='entry {}'.format(a)))
    event = OutputEvent(head=_event_head, entries=entries)
    return ParseResult(head=_parse_head, events=List(event))


class ParseSpec(TmuxCmdSpec):

    def custom(self):
        l1 = 'line 1'
        lines = List(l1, 'line 2', 'line 3')
        s = lines.mk_string('\\n')
        heads = List(_parse_head, _event_head)
        target = heads + lines / __.replace('line', 'entry')
        check2 = lambda: self.vim.buffer.content.should.equal(target)
        self.json_cmd('MyoShellCommand test', line="echo '{}'".format(s),
                      target='make',
                      parser='py:integration.tmux.parse_spec._parse_echo')
        self._open_pane('make')
        self.json_cmd('MyoRun test')
        self._output_contains(l1)
        self.json_cmd('MyoParse')
        later(check2)


class VimTestSpec(TmuxCmdSpec):

    @vimtest
    def vimtest(self):
        self.vim.vars.set_p('jump_to_error', False)
        self.vim.vars.set('test#python#runner', 'nose')
        fname = fixture_path('tmux', 'vim_test', 'test.py')
        target = str(fname.relative_to(base_dir().parent))
        def check1():
            self._output.exists(lambda a: target in a).should.be.true
        self.vim.cmd_sync('noswapfile edit {}'.format(fname))
        self.vim.buffer.vars.set_p('test_langs', ['python'])
        self.vim.cursor(5, 1)
        self.json_cmd('MyoVimTest')
        later(check1)
        target2 = 'RuntimeError: supernova'
        check2 = lambda: self.vim.buffer.content.last.should.contain(target2)
        self.json_cmd_sync('MyoParse')
        later(check2)
        self.vim.cursor(6, 1)
        self.vim.cmd_sync('call feedkeys("\\<cr>")')
        check3 = lambda: self.vim.buffer.name.should.equal(str(fname))
        later(check3)
        self.vim.focus(1).run_sync()
        self.vim.cmd_sync('call feedkeys("q")')
        check4 = lambda: self.vim.buffers.should.have.length_of(1)
        later(check4)

__all__ = ('ParseSpec',)



from integration._support.base import MyoIntegrationSpec

from amino.test.path import fixture_path
from amino import Map, Just, List
from amino.test import temp_file
from amino.lazy import lazy

from ribosome.test.integration import main_looped
from ribosome.machine.scratch import Mapping

from myo.command import Command
from myo.plugins.core.message import ParseOutput


class ParseHelpers:

    def _modifiable(self, val):
        self.vim.buffer.options('modifiable').should.contain(val)


class ParseSpec(MyoIntegrationSpec, ParseHelpers):

    @main_looped
    def native(self):
        cmd = Command(name='a', line='a', parser=Just('compiler'))
        fix = fixture_path('parse', 'err1')
        output = List.wrap(fix.read_text().splitlines())
        errorformat = List(
            '%AIn %f:%l',
            '%C  %.%#',
            '%+Z%.%#Error: %.%#'
        ).mk_string(',')
        self.vim.buffer.options.set('errorformat', errorformat)
        msg = ParseOutput(cmd, output, fix, Map())
        self.plugin.myo_start()
        self.root.send(msg)
        self._await()
        qf = self.vim.call('getqflist') | []
        qf.should_not.be.empty

    def twice(self):
        cmd = Command(name='a', line='a', langs=List('python'))
        msg = ParseOutput(cmd, List(), None, Map())
        self.plugin.myo_start()
        self.root.send_sync(msg)
        output_machine = self.root.sub[-1]
        self._modifiable(False)
        self.root.send_sync(Mapping(output_machine.uuid, 'q'))
        self._modifiable(True)
        self.root.send_sync(msg)
        self._modifiable(False)

_trace_len = 3
_line_count = _trace_len * 4 + 5


class ParseSpecBase(MyoIntegrationSpec):

    def _pre_start(self):
        super()._pre_start()
        self.vim.vars.set_p('jump_to_error', False)


class PythonParseSpec(ParseSpecBase):

    @lazy
    def _file(self):
        path = temp_file('core', 'parse', 'python', 'code.py')
        path.write_text(self._python_code.join_lines)
        return path

    @property
    def _python_func_name(self):
        return 'func_name'

    @lazy
    def _python_code(self):
        return List(
            'import functools',
            'def {}(a, b=1):'.format(self._python_func_name),
            '    return 1'
        )

    @property
    def _mk_trace(self):
        r = lambda: List.random_string(4)
        tmpl = '  File "{}", line 2, in {}'
        frame = lambda i: List(
            tmpl.format(self._file, self._python_func_name),
            '    {}()'.format(r())
        )
        t = List.range(_trace_len) // frame
        err = 'AttributeError: butt'
        return t.cat(err)

    def _init(self, parse_opt=Map()):
        trace1 = self._mk_trace
        trace2 = self._mk_trace
        output = trace1 + trace2
        cmd = Command(name='a', line='a', langs=List('python'))
        msg = ParseOutput(cmd, output, None, parse_opt)
        self.plugin.myo_start()
        self.root.send_sync(msg)
        return output

    def _go(self, line_count=_line_count, parse_opt=Map()):
        output = self._init(parse_opt)
        self.vim.buffer.content.should.contain(output[-1])
        self.vim.buffer.options('modifiable').should.contain(False)
        self.vim.buffer.content.should.have.length_of(line_count)

    def _check_jumped(self):
        self.vim.window.buffer.name.should.equal(str(self._file))
        self.vim.window.cursor.should.equal(List(2, 0))

    def jump(self):
        self._go()
        output_machine = self.root.sub[-1]
        self.vim.window.set_cursor(4)
        self.root.send_sync(Mapping(output_machine.uuid, '%cr%'))
        self.vim.windows.should.have.length_of(2)
        self._check_jumped()

    def filter(self):
        filters = List('py:integration.core.parse_spec._filter1')
        self._go(_trace_len * 2 + 3, Map(filters=filters))

    def initial_pos(self):
        self.vim.vars.set_p('first_error',
                            ['py:integration.core.parse_spec._first_error'])
        self._go()
        self.vim.window.line.should.contain(_line_count - 1)

    def initial_pos_with_jump(self):
        self.vim.vars.set_p('jump_to_error', True)
        self.vim.vars.set_p('first_error',
                            ['py:integration.core.parse_spec._first_error'])
        self._init()
        self._check_jumped()

    def default_jump(self):
        self.vim.vars.set_p('jump_to_error', True)
        self._init()
        self._check_jumped()


def _filter1(a):
    return a[:1] + a[1 + _trace_len * 2 + 2:]


def _first_error(a):
    return Just((_line_count - 1, 1))


class SbtParseSpec(ParseSpecBase):

    @lazy
    def _scala_file(self):
        path = temp_file('core', 'parse', 'sbt', 'code.scala')
        path.write_text(self._scala_code.join_lines)
        return path

    @property
    def _line(self):
        return 3

    @property
    def _scala_func_name(self):
        return 'scalafunc'

    @lazy
    def _scala_code(self):
        return List(
            'import foo.bar',
            'class Foo {',
            'def {}(a: Int): Int = a'.format(self._scala_func_name),
            '}',
        )

    @property
    def _mk_scala_trace(self):
        r = lambda: List.random_string(4)
        tmpl = '[error] {}:{}: butt'
        return List(
            tmpl.format(self._scala_file, self._line),
            '[error]    {}()'.format(r()),
            '[error]    ^',
        )

    def sbt(self):
        output = self._mk_scala_trace
        cmd = Command(name='a', line='a', langs=List('sbt'))
        msg = ParseOutput(cmd, output, None, Map())
        self.plugin.myo_start()
        self.root.send_sync(msg)
        output_machine = self.root.sub[-1]
        self.vim.buffer.content.should.contain(output[-1])
        self.vim.buffer.options('modifiable').should.contain(False)
        self.vim.window.set_cursor(4)
        self.root.send_sync(Mapping(output_machine.uuid, '%cr%'))
        self.vim.windows.should.have.length_of(2)
        self.vim.window.buffer.name.should.equal(str(self._scala_file))
        self.vim.window.cursor.should.equal(List(3, 4))

__all__ = ('ParseSpec',)



from collections import namedtuple

from ribosome.machine import may_handle, handle
from ribosome.machine.transition import Fatal

from myo.state import MyoComponent, MyoTransitions

from amino import L, _, List, Try, __, Maybe, curried, Map, Right, I
from myo.command import Command, VimCommand, ShellCommand, Shell
from myo.util import (optional_params, parse_callback_spec, amend_options,
                      bool_params)
from myo.plugins.core.message import Parse, ParseOutput, StageI
from myo.plugins.command.message import (Run, ShellRun, Dispatch, AddCommand,
                                         AddShellCommand, AddShell,
                                         AddVimCommand, SetShellTarget,
                                         CommandExecuted, RunTest, RunVimTest,
                                         CommandAdded, CommandShow, RunLatest)
from myo.plugins.command.util import assemble_vim_test_line

RunInShell = namedtuple('RunInShell', ['shell'])


class CommandTransitions(MyoTransitions):
    _test_cmd_name = '<test>'

    @property
    def _cmd_opt(self):
        return List('parser')

    @property
    def _cmd_list_opt(self):
        return List('langs', 'signals')

    @property
    def _cmd_bool_opt(self):
        return List('kill')

    def _create(self, tpe: type, mand_keys: List[str], opt_keys: List[str],
                **strict):
        o = self.msg.options
        opt = optional_params(o, *opt_keys + self._cmd_opt)
        opt_list = (optional_params(o, *self._cmd_list_opt)
                    .valmap(_ / List.wrap | List()))
        opt_bool = (bool_params(o, *self._cmd_bool_opt))
        missing = lambda: mand_keys.filter_not(o.has_key)
        err = lambda: 'cannot create {} without params: {}'.format(tpe,
                                                                   missing())
        mand = o.get_all_map(*mand_keys).to_either(err)
        return mand / (lambda a: tpe(**a, **opt, **opt_list, **opt_bool,
                                     **strict))

    def _add(self, tpe: type, mand_keys: List[str], opt_keys: List[str],
             **strict):
        return (
            self._create(tpe, mand_keys, opt_keys, **strict) /
            self._add_command
        )

    def _add_command(self, cmd: Command):
        return List(
            self.data.cat(cmd), CommandAdded(cmd, options=self.msg.options)
        )

    @may_handle(StageI)
    def stage_i(self):
        return AddShellCommand(name=self._test_cmd_name, options=Map(line=''))

    @handle(AddCommand)
    def add_command(self):
        return self._add(Command, List('line'), List(),
                         name=self.msg.name)

    @may_handle(AddVimCommand)
    def add_vim_command(self):
        return self._add(VimCommand, List('line'), self._cmd_opt,
                         name=self.msg.name)

    @handle(AddShellCommand)
    def add_shell_command(self):
        return self._add(ShellCommand, List('line'),
                         self._cmd_opt.cat('shell'), name=self.msg.name)

    @handle(AddShell)
    def add_shell(self):
        mand = List('line')
        opt = List('target')
        return self._create(Shell, mand, opt, name=self.msg.name) / (
            lambda shell: (
                self._add_command(shell) +
                (shell.target / L(SetShellTarget)(shell, _) / _.pub).to_list
            )
        )

    @may_handle(CommandAdded)
    def command_added(self):
        if self.msg.options.get('start').exists(I):
            return Run(self.msg.command.uuid)

    @handle(Run)
    def run(self):
        return (
            self.data.command(self.msg.command) /
            L(Dispatch)(_, self.msg.options)
        )

    @handle(ShellRun)
    def run_in_shell(self):
        line = self.msg.options.get('line') | ''
        def send(cmd):
            return (
                self.data.transient_command(cmd),
                Dispatch(cmd, self.msg.options)
            )
        # double Maybe is necessary - the outer indicates success, the
        # inner is the type of the ShellCommand field
        return (
            self.data.shell(self.msg.shell) /
            _.uuid /
            Maybe /
            (lambda a: ShellCommand(name='shell_run', shell=a, line=line,
                                    transient=True)) /
            send
        )

    @handle(RunLatest)
    def run_latest(self):
        return (
            self.data.commands.latest_command /
            Dispatch
        )

    @handle(Dispatch)
    def dispatch(self):
        cmd = self.msg.command
        self.vim.vars.set_p('last_command', dict(name=cmd.name))
        self.vim.pautocmd('RunCommand')
        return (
            self.data.dispatch_message(cmd, self.msg.options) /
            _.pub /
            (lambda a: (a, CommandExecuted(cmd)))
        )

    @may_handle(CommandExecuted)
    def command_executed(self):
        new = self.data.commands.add_history(self.msg.command.uuid)
        return self.data.set(commands=new)

    @handle(Parse)
    def parse(self):
        cmd = (
            self.msg.options.get('command')
            .cata(self.data.command, lambda: self.data.commands.latest_command)
            .to_either('invalid command name or empty history')
        )
        def find_log(c):
            return (
                (c.shell // self.data.shell | c)
                .log_path.to_either('command has no log')
            )
        log_path = cmd // find_log
        def parse(c, l):
            return (
                Try(l.read_text) /
                __.splitlines() /
                List.wrap /
                L(ParseOutput)(c, _, l, self.msg.options) /
                _.pub
            )
        return (cmd & log_path).flat_map2(parse)

    @handle(RunTest)
    def run_test(self):
        opt = self.msg.options
        return (
            opt.get('ctor').to_either(Fatal('no test ctor specified')) //
            self._assemble //
            self._run_test_line(opt - 'ctor')
        ).lmap(Fatal)

    @handle(RunVimTest)
    def run_vim_test(self):
        return (
            assemble_vim_test_line(self.vim) //
            __.head.to_either('vim-test failed') //
            self._run_test_line(self.msg.options)
        )

    @may_handle(CommandShow)
    def show(self):
        msg = self.data.commands.commands / _.desc
        self.log.info(msg.join_lines)

    def _assemble(self, ctor):
        return parse_callback_spec(ctor).join / (lambda a: a())

    @curried
    def _run_test_line(self, options, line):
        glangs = self.vim.buffer.vars.pl('test_langs')
        langs = options.get('langs').or_else(glangs) | List()
        shell = self.vim.buffer.pvar_or_global('test_shell')
        target = self.vim.buffer.pvar_or_global('test_target')
        opt = amend_options(options, 'shell', shell)
        opt2 = amend_options(opt, 'target', target)
        def dispatch(data):
            return Right(data) & (data.command(self._test_cmd_name) /
                                  L(Dispatch)(_, opt2))
        return (
            self.data.command_lens(self._test_cmd_name)
            .to_either('no test command') /
            __.modify(__.set(line=line, langs=langs)) //
            dispatch
        )


class Plugin(MyoComponent):
    Transitions = CommandTransitions

    def new_state(self):
        pass

__all__ = ('Plugin', 'AddVimCommand', 'AddCommand', 'Run', 'AddShell',
           'AddShellCommand', 'SetShellTarget')



from uuid import UUID

from ribosome.record import Record as RecordBase, uuid_field, field

from myo.ui.tmux.util import Ident


class Record(RecordBase):
    uuid = uuid_field()

    def __str__(self):
        return super().__repr__()

    @property
    def ident(self):
        return self.uuid


class Named(Record):
    name = field(str)

    def has_ident(self, ident: Ident):
        attr = self.uuid if isinstance(ident, UUID) else self.name
        return attr == ident

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, self.name)

__all__ = ('Record',)



from operator import add

from ribosome.machine import message, may_handle, Machine, handle, Nop
from ribosome.nvim import ScratchBuffer, NvimFacade
from ribosome.machine.scratch import ScratchMachine, Quit
from ribosome.machine.base import UnitTask
from ribosome.machine.state import Init
from ribosome.record import field

from amino.task import Task
from amino import Map, _, L, Left, __, List, Either, Just
from amino.lazy import lazy

from myo.output.data import ParseResult, Location
from myo.state import MyoTransitions
from myo.logging import Logging
from myo.record import Record
from myo.util import parse_callback_spec

Jump = message('Jump')
FirstError = message('FirstError')
DisplayLines = message('DisplayLines')


class ResultAdapter(Logging):

    def __init__(self, vim, result: ParseResult, filters: List[str]) -> None:
        self.vim = vim
        self.result = result
        self.filters = filters

    @property
    def filter_callbacks(self):
        return (
            self.filters / parse_callback_spec / __.right_or_map(Left)
        ).sequence(Either)

    @lazy
    def lines(self):
        def fold(cbs):
            return cbs.fold_left(self.result.lines)(lambda z, a: a(z))
        return self.filter_callbacks / fold

    @property
    def display_lines(self):
        return (self.lines.eff() / _.text).value

    def target_for_line(self, line):
        return self.lines // __.lift(line) / _.target

    @property
    def langs(self):
        return self.result.langs

    @property
    def events(self):
        return self.result.events


class OMState(Record):
    result = field(ResultAdapter)


class OutputMachineTransitions(MyoTransitions):

    @property
    def scratch(self):
        return self.machine.scratch

    @property
    def buffer(self):
        return self.scratch.buffer

    @property
    def result(self):
        return self.machine.result

    @property
    def _special_jumps(self):
        return Map(default=self._jump_default, first=self._jump_first,
                   last=self._jump_last)

    @property
    def _lang_jumps(self):
        return Map(python=self._jump_last)

    @may_handle(Init)
    def init(self):
        return self.with_sub(self.state), DisplayLines()

    @handle(DisplayLines)
    def display_lines(self):
        def run(lines):
            return (
                Task.call(self.buffer.set_content, lines) &
                Task.call(self.buffer.set_modifiable, False)
            ).replace(Just(FirstError()))
        return self.result.display_lines / run

    @handle(FirstError)
    def first_error(self):
        spec = (
            self.machine.options.get('first_error')
            .o(self.vim.vars.pl('first_error')) |
            List()
        ).cat('default')
        special, custom_spec = spec.split(self._special_jumps.contains)
        special_cb = special // self._special_jumps.get
        jump = self.vim.vars.pb('jump_to_error') | True
        next_step = Jump() if jump else Nop()
        return (
            (
                custom_spec /
                parse_callback_spec /
                _.ljoin
            ).sequence(Either) // (
                __
                .add(special_cb)
                .find_map(__(self.result))
                .map2(L(Task.call)(self.vim.window.set_cursor, _, _))
                .map(__.replace(Just(next_step)))
            )
        )

    @handle(Jump)
    def jump(self):
        target = (
            self.vim.window.line /
            (_ - 1) //
            self.result.target_for_line
        ).filter_type(Location).to_either('not a location')
        open_file = target / L(self._open_file)(_.file_path)
        set_line = (
            (target / _.coords)
            .map2(lambda a, b: Task(lambda: self.vim.window.set_cursor(a, b)))
        )
        return (open_file & set_line).map2(add) / UnitTask

    @property
    def _jump_default(self):
        return (self.result.langs.find_map(self._lang_jumps.get) |
                (lambda: self._jump_first))

    def _position_index(self, lines):
        return (
            lines
            .index_where(lambda a: isinstance(a.target, Location))
            .map(lambda a: (a + 1, 1))
        )

    def _jump_first(self, result):
        return (self.result.lines // self._position_index)

    def _jump_last(self, result):
        def run(lines):
            return (
                self._position_index(lines)
                .map2(lambda l, c: (lines.length - l, c))
            )
        return self.result.lines / _.reversed // run

    def _open_file(self, path):
        def split():
            # TODO
            pass
        if not path.is_file():
            return Task.now(Left('not a file: {}'.format(path)))
        else:
            win = (
                self.vim.windows.find(__.buffer.modifiable.contains(True))
                .or_else(split)
                .task('could not get a window') /
                __.focus()
            )
            edit = Task.call(self.vim.edit, path) / __.run_async()
            return win + edit


class OutputMachine(ScratchMachine, Logging):
    Transitions = OutputMachineTransitions

    def __init__(self, vim: NvimFacade, scratch: ScratchBuffer,
                 result: ResultAdapter, parent: Machine, options: Map) -> None:
        super().__init__(vim, scratch, parent=parent, title='output')
        self.result = result
        self.options = options

    @property
    def prefix(self):
        return 'Myo'

    @property
    def mappings(self):
        return Map({
            '%cr%': Jump,
            'q': Quit,
        })

    def new_state(self):
        return OMState(result=self.result)

__all__ = ('OutputMachine',)



from typing import Callable

from myo.ui.tmux.view import View
from myo.ui.tmux.pane import Pane
from myo.ui.tmux.util import Ident

from lenses import lens

from amino import _, __, List
from amino.lazy import lazy

from ribosome.record import dfield, list_field, field, Record


class LayoutDirection(Record):
    id = field(str)


class LayoutDirections:
    vertical = LayoutDirection(id='vertical')
    horizontal = LayoutDirection(id='horizontal')
    all = List(vertical, horizontal)
    all_ids = all / _.id  # type: List[str]

    @staticmethod
    def parse(s: str):
        return (LayoutDirections.horizontal if s == 'horizontal' else
                LayoutDirections.vertical)


class Layout(View):
    panes = list_field()
    layouts = list_field()
    direction = dfield(LayoutDirections.vertical)

    def pane_index(self, f: Callable[[Pane], bool]):
        self.panes.index_where(f)

    def find_pane_pred(self, pred: Callable[[Pane], bool]):
        return (self.panes.find(pred)
                .or_else(lambda: self._find_pane_in_layouts(pred)))

    def _find_pane_in_layouts(self, pred: Callable[[Pane], bool]):
        return self.layouts.find_map(__.find_pane_pred(pred))

    def find_pane(self, ident: Ident):
        return self.find_pane_pred(__.has_ident(ident))

    @staticmethod
    def pane_lens(root, f: Callable[[Pane], bool]):
        def g(l):
            return (
                l.panes.lens(f)
                .map(lens().panes.add_lens)
                .or_else(lambda: h(l))
            )
        def h(l):
            return l.layouts.deep_lens(g) / lens().layouts.add_lens
        return h(root) / lens(root).add_lens

    @lazy
    def views(self):
        return self.panes + self.layouts

    @lazy
    def horizontal(self):
        return self.direction == LayoutDirections.horizontal

    @lazy
    def weights(self):
        w = self.views / _.weight
        total = sum(w.join)
        return w / (lambda a: a / (_ / total))

    @lazy
    def actual_min_sizes(self):
        return self.views / (lambda a:
                             a.effective_fixed_size.or_else(a.min_size) | 0)

    @lazy
    def actual_max_sizes(self):
        return self.views / (lambda a:
                             a.effective_fixed_size.or_else(a.max_size) | 999)

    @property
    def desc(self):
        dir = 'H' if self.horizontal else 'V'
        return 'L {} \'{}\' {}'.format(dir, self.name, self.size_desc)

    @property
    def all_panes(self):
        return self.panes + (self.layouts // _.all_panes)


class VimLayout(Layout):

    @property
    def desc(self):
        return 'V{}'.format(super().desc)

__all__ = ('Layout', 'LayoutDirection', 'VimLayout')



from functools import singledispatch  # type: ignore

from ribosome.machine import Error

from myo.dispatch import Dispatcher
from myo.command import ShellCommand, Shell
from myo.plugins.tmux.message import (TmuxRunCommand, TmuxRunLineInShell,
                                       TmuxRunShell)
from myo.plugins.command.main import RunInShell


@singledispatch
def convert_message(msg, options):
    return Error('invalid message for tmux dispatch: {}'.format(msg))


@convert_message.register(RunInShell)
def _convert_run_in_shell(cmd, options):
    return TmuxRunLineInShell(cmd.shell, options)


@convert_message.register(Shell)  # type: ignore
def _convert_shell(cmd, options):
    return TmuxRunShell(cmd, options)


@convert_message.register(ShellCommand)  # type: ignore
def _convert_shell_command(cmd, options):
    return TmuxRunCommand(cmd, options)


class TmuxDispatcher(Dispatcher):

    def can_run(self, cmd):
        return isinstance(cmd, (ShellCommand, RunInShell))

    def message(self, cmd, options):
        return convert_message(cmd, options)

__all__ = ('TmuxDispatcher',)



from typing import Callable, Any

from amino.task import Task

from lenses import Lens, lens
from amino import Either, List, F, _, L, Left, __, Right, I
from amino.lens.tree import path_lens_unbound_pre

from ribosome.record import Record, field, list_field
from ribosome.machine import Message
from ribosome.machine.transition import Error

from myo.ui.tmux.window import Window
from myo.logging import Logging
from myo.ui.tmux.pane import Pane
from myo.ui.tmux.layout import Layout
from myo.ui.tmux.view import View


class PanePath(Record):
    window = field(Window)
    pane = field(Pane)
    layout = field(Layout)
    outer = list_field()

    @staticmethod
    def create(window, pane, layout, outer):
        return PanePath(window=window, pane=pane, layout=layout, outer=outer)

    @staticmethod
    def try_create(views: List[View]) -> Either[str, 'PanePath']:
        def try_create2(pane, win, layouts):
            return (
                layouts.detach_last.map2(F(PanePath.create, win, pane))
                .to_either('PanePath.create: last item is not Pane')
            )
        def try_create1(pane, layouts):
            return (
                layouts.detach_head
                .to_either('PanePath.create: last item is not Pane')
                .flat_map2(F(try_create2, pane))
            )
        return (
            views.detach_last.to_either('PanePath.create: empty List[View]')
            .flat_map2(try_create1)
        )

    @property
    def layouts(self):
        return self.outer.cat(self.layout)

    @property
    def to_list(self):
        return self.layouts.cat(self.pane).cons(self.window)

    def map(self, fp: Callable[[Pane], Pane], fl: Callable[[Layout], Layout]):
        return PanePath.create(
            self.window, fp(self.pane), fl(self.layout), self.outer / fl)

    def map_pane(self, f: Callable[[Pane], Pane]):
        return self.map(f, I)

PPTrans = Callable[[PanePath], Task[Either[Any, PanePath]]]


class PanePathLens(Record):
    lens = field(Lens)

    @staticmethod
    def create(lens: Lens):
        return PanePathLens(lens=lens)

    def run(self, win: Window, f: PPTrans) -> Task[Either[Any, Window]]:
        bound = self.lens.bind(win)
        path = PanePath.try_create(List.wrap(bound.get()))
        def update(pp):
            return bound.set(pp.to_list)
        def apply(pp):
            return f(pp) / (_ / update)
        return Task.from_either(path) // apply


def _initial_ppm_f(pp, window) -> Task[Either[Any, Window]]:
    return Task.now(Right(window))


class PanePathMod(Logging, Message):
    pred = field(Callable)
    _f = field(Callable, initial=(lambda: _initial_ppm_f))

    def _chain(self, f: PPTrans, g: Callable):
        h = lambda pp: L(pp.run)(_, f)
        chain = lambda pp, win: self._f(pp, win) // g(win, h(pp))
        return PanePathMod(pred=self.pred, _f=chain)

    def map(self, f: PPTrans) -> Task[Either[Any, Window]]:
        keep_error = F(Left) >> Task.now
        return self._chain(f, lambda w, d: __.cata(keep_error, d))

    __truediv__ = map

    def and_then(self, f: PPTrans):
        return self._chain(f, lambda w, d: lambda a: d(a | w))

    __add__ = and_then

    def run(self, window: Window) -> Task:
        return self._lens(window) // L(self._f)(_, window)

    def _lens(self, window):
        f = __.panes.find_lens_pred(self.pred).map(lens().panes.add_lens)
        sub = _.layouts
        pre = _.root
        return (
            Task.from_maybe(
                path_lens_unbound_pre(window, sub, f, pre),
                Error('lens path failed for {}'.format(self.pred))
            ) /
            PanePathLens.create
        )


def ppm_id(path: PanePathMod):
    return Task.now(Right(path))

__all__ = ('PanePath', 'PanePathLens', 'PanePathMod', 'ppm_id')



from myo.dispatch import Dispatcher
from myo.command import VimCommand
from myo.plugins.core.message import RunVimCommand

from ribosome.nvim import HasNvim


class VimDispatcher(Dispatcher, HasNvim):

    def can_run(self, cmd):
        return isinstance(cmd, VimCommand)

    def message(self, cmd, options):
        return RunVimCommand(cmd, options)

__all__ = ('VimDispatcher',)
